@@ -1,4 +1,5 @@
 import enum
+import abc
 
 from . import util
 
@@ -23,11 +24,9 @@ class GameMode (util.Typed, util.Identified):
 
 
 class GameModes (enum.Enum):
-    OPEN_WORLD = GameMode('Open World', ('open world', 'dungeons'))
+    OPEN_WORLD = GameMode('Open World', ('open world', 'dungeons', 'low-tier fractals'))
     DUNGEONS = GameMode('Dungeons', ('dungeons',))
-    FRACTALS = GameMode('Fractals',
-                         ('unorganised fractals', 'dungeons', 'open world'))
-    RAIDS = GameMode('Raids', ('casual raids', 'casual organised fractals'))
+    RAIDS = GameMode('Raids', ('raids', 'fractals'))
     PVP = GameMode('PvP', ('PvP',))
     WVW = GameMode('WvW', ('WvW',))
 
@@ -38,7 +37,9 @@ class GameModes (enum.Enum):
 _game_modes_id_lookup = _enum_id_lookup(GameModes)
 
 
-class BuildMetadata (util.Typed):
+class BuildMetadata (util.Record, util.Typed):
+    _attrs = ('game_mode', 'profession', 'elite_spec', 'labels')
+
     def __init__ (self, game_mode, profession, elite_spec, labels=()):
         self.game_mode = game_mode # can be None
         self.profession = profession
@@ -60,12 +61,11 @@ class BuildMetadata (util.Typed):
 
 
 class TextBody:
-    pass
+    def __init__ (self, source):
+        self.source = source
 
-
-class MarkdownBody (TextBody):
-    def __init__ (self, text):
-        self.text = text
+    def __repr__ (self):
+        return self.source
 
 
 class WeaponType (util.Typed, util.Identified):
@@ -75,11 +75,11 @@ class WeaponType (util.Typed, util.Identified):
 
 
 class WeaponTypes (enum.Enum):
-    GREATSWORD = WeaponType('greatsword', 2)
+    GREATSWORD = WeaponType(('greatsword', 'gs'), 2)
     HAMMER = WeaponType('hammer', 2)
-    LONGBOW = WeaponType('longbow', 2)
+    LONGBOW = WeaponType(('longbow', 'lb'), 2)
     RIFLE = WeaponType('rifle', 2)
-    SHORTBOW = WeaponType('shortbow', 2)
+    SHORTBOW = WeaponType(('shortbow', 'sb'), 2)
     STAFF = WeaponType('staff', 2)
 
     AXE = WeaponType('axe', 1)
@@ -91,7 +91,7 @@ class WeaponTypes (enum.Enum):
     FOCUS = WeaponType('focus', 1)
     SHIELD = WeaponType('shield', 1)
     TORCH = WeaponType('torch', 1)
-    WARHORN = WeaponType('warhorn', 1)
+    WARHORN = WeaponType(('warhorn', 'wh'), 1)
 
     # disallowed until aquatic weapons are properly supported
     #HARPOON = WeaponType(('harpoon', 'speargun'), 2)
@@ -169,7 +169,9 @@ class PvpSigils (enum.Enum):
 _pvp_sigils_id_lookup = _enum_id_lookup(PvpSigils)
 
 
-class Weapon (util.Typed):
+class Weapon (util.Record, util.Typed):
+    _attrs = ('type_', 'hand', 'stats', 'sigils')
+
     def __init__ (self, type_, hand, stats, sigils):
         self.type_ = type_
         self.hand = hand
@@ -188,7 +190,9 @@ class Weapon (util.Typed):
         return (self.type_, self.hand, self.stats, self.sigils)
 
 
-class Weapons (util.Typed):
+class Weapons (util.Record, util.Typed):
+    _attrs = ('set1', 'set2')
+
     def __init__ (self, set1, set2=None):
         self.set1 = tuple(set1)
         self.set2 = None if set2 is None else tuple(set2)
@@ -335,7 +339,9 @@ class ArmourTypes (enum.Enum):
 _armour_types_id_lookup = _enum_id_lookup(ArmourTypes)
 
 
-class ArmourPiece (util.Typed):
+class ArmourPiece (util.Record, util.Typed):
+    _attrs = ('type_', 'stats', 'rune')
+
     def __init__ (self, type_, stats, rune):
         self.type_ = type_
         self.stats = stats
@@ -348,8 +354,12 @@ class ArmourPiece (util.Typed):
         return (self.type_, self.stats, self.rune)
 
 
-class Armour (util.Typed):
+class Armour (util.Record, util.Typed):
+    _attrs = ('pieces',)
+
     def __init__ (self, pieces):
+        if isinstance(pieces, dict):
+            pieces = pieces.values()
         self.pieces = {p.type_: p for p in pieces}
 
         if len(pieces) != 6:
@@ -363,7 +373,9 @@ class Armour (util.Typed):
         return tuple(self.pieces.items())
 
 
-class PvpArmour (util.Typed):
+class PvpArmour (util.Record, util.Typed):
+    _attrs = ('rune',)
+
     def __init__ (self, rune):
         self.rune = rune
 
@@ -383,7 +395,9 @@ class TrinketTypes (enum.Enum):
     RING_2 = 'ring 2'
 
 
-class Trinket (util.Typed):
+class Trinket (util.Record, util.Typed):
+    _attrs = ('type_', 'stats')
+
     def __init__ (self, type_, stats):
         self.type_ = type_
         self.stats = stats
@@ -395,8 +409,12 @@ class Trinket (util.Typed):
         return (self.type_, self.stats)
 
 
-class Trinkets (util.Typed):
+class Trinkets (util.Record, util.Typed):
+    _attrs = ('pieces',)
+
     def __init__ (self, pieces):
+        if isinstance(pieces, dict):
+            pieces = pieces.values()
         self.pieces = {p.type_: p for p in pieces}
 
         if len(pieces) != 6:
@@ -432,7 +450,9 @@ _gear_groups_id_lookup.update(_armour_types_id_lookup)
 _gear_groups_id_lookup.update({type_.value: type_ for type_ in TrinketTypes})
 
 
-class Consumables (util.Typed):
+class Consumables (util.Record, util.Typed):
+    _attrs = ('food', 'utility')
+
     def __init__ (self, food=None, utility=None):
         self.food = food
         self.utility = utility
@@ -441,7 +461,9 @@ class Consumables (util.Typed):
         return (self.food, self.utility)
 
 
-class Gear (util.Typed):
+class Gear (util.Record, util.Typed):
+    _attrs = ('weapons', 'armour', 'trinkets', 'consumables')
+
     def __init__ (self, weapons, armour, trinkets, consumables):
         self.weapons = weapons
         self.armour = armour
@@ -455,7 +477,9 @@ class Gear (util.Typed):
         self.weapons.check_profession(profession, elite_spec)
 
 
-class PvpGear (util.Typed):
+class PvpGear (util.Record, util.Typed):
+    _attrs = ('stats', 'weapons', 'armour')
+
     def __init__ (self, stats, weapons, armour):
         self.stats = stats
         self.weapons = weapons
@@ -468,9 +492,53 @@ class PvpGear (util.Typed):
         self.weapons.check_profession(profession, elite_spec)
 
 
+class TraitTier (util.Typed):
+    def __init__ (self, api_id, name):
+        self.api_id = api_id
+        self.index = None if api_id == 0 else (api_id - 1)
+        self.name = name
+
+    def _value (self):
+        return self.api_id
+
+
+class TraitTiers (enum.Enum):
+    WEAPON = TraitTier(0, 'Weapon')
+    ADEPT = TraitTier(1, 'Adept')
+    MASTER = TraitTier(2, 'Master')
+    GRANDMASTER = TraitTier(3, 'Grandmaster')
+
+    @staticmethod
+    def from_api_id (api_id):
+        return _trait_tiers_api_id_lookup[api_id]
+
+_trait_tiers_api_id_lookup = {tier.value.api_id: tier for tier in TraitTiers}
+
+
+class TraitType (util.Typed):
+    def __init__ (self, api_id):
+        self.api_id = api_id
+        self.name = api_id
+
+    def _value (self):
+        return self.api_id
+
+
+class TraitTypes (enum.Enum):
+    MINOR = TraitType('Minor')
+    MAJOR = TraitType('Major')
+
+    @staticmethod
+    def from_api_id (api_id):
+        return _trait_types_api_id_lookup[api_id]
+
+_trait_types_api_id_lookup = {type_.value.api_id: type_ for type_ in TraitTypes}
+
+
 class TraitChoice (util.Typed):
     def __init__ (self, index, name):
         self.index = index
+        self.api_id = index
         self.name = name
 
     def _value (self):
@@ -486,11 +554,17 @@ class TraitChoices (enum.Enum):
     def from_index (index):
         return _trait_choices_index_lookup[index]
 
+    @staticmethod
+    def from_api_id (api_id):
+        return TraitChoices.from_index(api_id)
+
 _trait_choices_index_lookup = {choice.value.index: choice
                                for choice in TraitChoices}
 
 
-class SpecialisationChoices (util.Typed):
+class SpecialisationChoices (util.Record, util.Typed):
+    _attrs = ('spec', 'choices')
+
     def __init__ (self, spec, choices):
         self.spec = spec
         self.choices = tuple(choices) # items can be None
@@ -507,7 +581,9 @@ class SpecialisationChoices (util.Typed):
         return (self.spec, self.choices)
 
 
-class Traits (util.Typed):
+class Traits (util.Record, util.Typed):
+    _attrs = ('specs',)
+
     def __init__ (self, specs):
         self.specs = tuple(specs) # items can be None
 
@@ -577,7 +653,9 @@ class SkillTypes (enum.Enum):
 _skill_types_id_lookup = _enum_id_lookup(SkillTypes)
 
 
-class Skills (util.Typed):
+class Skills (util.Record, util.Typed):
+    _attrs = ('heal', 'utilities', 'elite')
+
     def __init__ (self, heal, utilities, elite):
         # skills can be None
         self.heal = heal
@@ -609,9 +687,7 @@ class Skills (util.Typed):
             if skill is None:
                 continue
 
-            if (len(skill.professions) != 1 or
-                next(iter(skill.professions)) != profession
-            ):
+            if skill.profession is not None and skill.profession != profession:
                 raise BuildError(f'not a skill for {profession.name}: '
                                  f'{skill.name}')
 
@@ -625,7 +701,9 @@ class Skills (util.Typed):
                 raise BuildError(f'skill not usable underwater: {skill.name}')
 
 
-class RevenantSkills (util.Typed):
+class RevenantSkills (util.Record, util.Typed):
+    _attrs = ('legends',)
+
     def __init__ (self, legends):
         self.legends = tuple(legends) # items can be None
 
@@ -645,7 +723,9 @@ class RevenantSkills (util.Typed):
                 raise BuildError(f'legend not usable underwater: {legend.name}')
 
 
-class RangerPets (util.Typed):
+class RangerPets (util.Record, util.Typed):
+    _attrs = ('pets',)
+
     def __init__ (self, pets):
         self.pets = tuple(pets) # items can be None
 
@@ -656,7 +736,9 @@ class RangerPets (util.Typed):
         return self.pets
 
 
-class RangerOptions (util.Typed):
+class RangerOptions (util.Record, util.Typed):
+    _attrs = ('pets', 'aquatic_pets')
+
     def __init__ (self, pets, aquatic_pets=None):
         self.pets = pets
         self.aquatic_pets = (RangerPets((None, None))
@@ -666,7 +748,10 @@ class RangerOptions (util.Typed):
         return (self.pets, self.aquatic_pets)
 
 
-class Intro:
+class Intro (util.Record):
+    _attrs = ('url', 'description', 'gear', 'traits', 'skills',
+              'profession_options', 'aquatic_skills')
+
     def __init__ (self, url, description, gear, traits, skills,
                   profession_options=None, aquatic_skills=None):
         self.url = url # can be None
@@ -736,7 +821,9 @@ class BoonTargets (enum.Enum):
 _boon_targets_id_lookup = _enum_id_lookup(BoonTargets)
 
 
-class BoonUptime:
+class BoonUptime (util.Record):
+    _attrs = ('boon', 'target', 'uptime_percent')
+
     def __init__ (self, boon, target, uptime_percent):
         self.boon = boon
         self.target = target
@@ -753,7 +840,10 @@ class BoonNotes:
         self.boon_uptime_variants = boon_uptime_variants
 
 
-class Build:
+class Build (util.Record):
+    _attrs = ('metadata', 'intro', 'alternatives', 'usage', 'notes',
+              'boon_notes', 'encounters')
+
     def __init__ (self, metadata, intro, alternatives=None, usage=None,
                   notes=None, boon_notes=None, encounters=None):
         self.metadata = metadata
